@@ -26,7 +26,11 @@ module Data.MonoTraversable where
 
 import           Control.Applicative
 import           Control.Category
+#if MIN_VERSION_base(4,8,0)
+import           Control.Monad        (Monad (..))
+#else
 import           Control.Monad        (Monad (..), liftM)
+#endif
 import qualified Data.ByteString      as S
 import qualified Data.ByteString.Lazy as L
 import qualified Data.Foldable        as F
@@ -36,13 +40,12 @@ import           Data.Monoid (Monoid (..), Any (..), All (..))
 import qualified Data.Text            as T
 import qualified Data.Text.Lazy       as TL
 import           Data.Traversable
-import           Data.Traversable.Instances ()
 import           Data.Word            (Word8)
 import Data.Int (Int, Int64)
 import           GHC.Exts             (build)
 import           Prelude              (Bool (..), const, Char, flip, IO, Maybe (..), Either (..),
                                        (+), Integral, Ordering (..), compare, fromIntegral, Num, (>=),
-                                       seq, otherwise, Eq, Ord, (-), (*))
+                                       (==), seq, otherwise, Eq, Ord, (-), (*))
 import qualified Prelude
 import qualified Data.ByteString.Internal as Unsafe
 import qualified Foreign.ForeignPtr.Unsafe as Unsafe
@@ -63,15 +66,6 @@ import Data.HashMap.Strict (HashMap)
 import Data.Vector (Vector)
 import Control.Monad.Trans.Maybe (MaybeT (..))
 import Control.Monad.Trans.List (ListT)
-import Control.Monad.Trans.Identity (IdentityT)
-import Data.Functor.Apply (MaybeApply (..), WrappedApplicative)
-import Control.Comonad (Cokleisli, Comonad, extract, extend)
-import Control.Comonad.Store (StoreT)
-import Control.Comonad.Env (EnvT)
-import Control.Comonad.Traced (TracedT)
-#if !MIN_VERSION_comonad(5,0,0)
-import Data.Functor.Coproduct (Coproduct)
-#endif
 import Control.Monad.Trans.Writer (WriterT)
 import qualified Control.Monad.Trans.Writer.Strict as Strict (WriterT)
 import Control.Monad.Trans.State (StateT(..))
@@ -79,11 +73,9 @@ import qualified Control.Monad.Trans.State.Strict as Strict (StateT(..))
 import Control.Monad.Trans.RWS (RWST(..))
 import qualified Control.Monad.Trans.RWS.Strict as Strict (RWST(..))
 import Control.Monad.Trans.Reader (ReaderT)
-import Control.Monad.Trans.Error (ErrorT(..))
 import Control.Monad.Trans.Cont (ContT)
 import Data.Functor.Compose (Compose)
 import Data.Functor.Product (Product)
-import Data.Semigroupoid.Static (Static)
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.HashSet (HashSet)
@@ -95,8 +87,7 @@ import qualified Data.Vector.Storable as VS
 import qualified Data.IntSet as IntSet
 import Data.Semigroup (Semigroup, Option (..), Arg)
 import qualified Data.ByteString.Unsafe as SU
-import Data.DList (DList)
-import qualified Data.DList as DL
+import Control.Monad.Trans.Identity (IdentityT)
 
 -- | Type family for getting the type of the elements
 -- of a monomorphic container.
@@ -111,7 +102,6 @@ type instance Element (ZipList a) = a
 type instance Element (Maybe a) = a
 type instance Element (Tree a) = a
 type instance Element (Seq a) = a
-type instance Element (DList a) = a
 type instance Element (ViewL a) = a
 type instance Element (ViewR a) = a
 type instance Element (IntMap a) = a
@@ -130,9 +120,6 @@ type instance Element (Set e) = e
 type instance Element (HashSet e) = e
 type instance Element (Vector a) = a
 type instance Element (WrappedArrow a b c) = c
-type instance Element (MaybeApply f a) = a
-type instance Element (WrappedApplicative f a) = a
-type instance Element (Cokleisli w a b) = b
 type instance Element (MaybeT m a) = a
 type instance Element (ListT m a) = a
 type instance Element (IdentityT m a) = a
@@ -143,20 +130,12 @@ type instance Element (Strict.StateT s m a) = a
 type instance Element (RWST r w s m a) = a
 type instance Element (Strict.RWST r w s m a) = a
 type instance Element (ReaderT r m a) = a
-type instance Element (ErrorT e m a) = a
 type instance Element (ContT r m a) = a
 type instance Element (Compose f g a) = a
 type instance Element (Product f g a) = a
-type instance Element (Static f a b) = b
 type instance Element (U.Vector a) = a
 type instance Element (VS.Vector a) = a
 type instance Element (Arg a b) = b
-type instance Element (EnvT e w a) = a
-type instance Element (StoreT s w a) = a
-type instance Element (TracedT m w a) = a
-#if !MIN_VERSION_comonad(5,0,0)
-type instance Element (Coproduct f g a) = a
-#endif
 
 -- | Monomorphic containers that can be mapped over.
 class MonoFunctor mono where
@@ -184,7 +163,6 @@ instance MonoFunctor (ZipList a)
 instance MonoFunctor (Maybe a)
 instance MonoFunctor (Tree a)
 instance MonoFunctor (Seq a)
-instance MonoFunctor (DList a)
 instance MonoFunctor (ViewL a)
 instance MonoFunctor (ViewR a)
 instance MonoFunctor (IntMap a)
@@ -200,16 +178,7 @@ instance MonoFunctor (Map k v)
 instance MonoFunctor (HashMap k v)
 instance MonoFunctor (Vector a)
 instance MonoFunctor (Arg a b)
-instance Functor w => MonoFunctor (EnvT e w a)
-instance Functor w => MonoFunctor (StoreT s w a)
-instance Functor w => MonoFunctor (TracedT m w a)
-#if !MIN_VERSION_comonad(5,0,0)
-instance (Functor f, Functor g) => MonoFunctor (Coproduct f g a)
-#endif
 instance Arrow a => MonoFunctor (WrappedArrow a b c)
-instance Functor f => MonoFunctor (MaybeApply f a)
-instance Functor f => MonoFunctor (WrappedApplicative f a)
-instance MonoFunctor (Cokleisli w a b)
 instance Functor m => MonoFunctor (MaybeT m a)
 instance Functor m => MonoFunctor (ListT m a)
 instance Functor m => MonoFunctor (IdentityT m a)
@@ -220,11 +189,9 @@ instance Functor m => MonoFunctor (Strict.StateT s m a)
 instance Functor m => MonoFunctor (RWST r w s m a)
 instance Functor m => MonoFunctor (Strict.RWST r w s m a)
 instance Functor m => MonoFunctor (ReaderT r m a)
-instance Functor m => MonoFunctor (ErrorT e m a)
 instance Functor m => MonoFunctor (ContT r m a)
 instance (Functor f, Functor g) => MonoFunctor (Compose f g a)
 instance (Functor f, Functor g) => MonoFunctor (Product f g a)
-instance Functor f => MonoFunctor (Static f a b)
 instance U.Unbox a => MonoFunctor (U.Vector a) where
     omap = U.map
     {-# INLINE omap #-}
@@ -232,6 +199,21 @@ instance VS.Storable a => MonoFunctor (VS.Vector a) where
     omap = VS.map
     {-# INLINE omap #-}
 
+-- | @'replaceElem' old new@ replaces all @old@ elements with @new@.
+--
+-- @since 1.0.1
+replaceElem :: (MonoFunctor mono, Eq (Element mono)) => Element mono -> Element mono -> mono -> mono
+replaceElem old new = omap (\x -> if x == old then new else x)
+
+{-# INLINE [0] replaceElem #-}
+{-# RULES "strict Text replaceElem" replaceElem = replaceElemStrictText #-}
+replaceElemStrictText :: Char -> Char -> T.Text -> T.Text
+replaceElemStrictText old new = T.replace (T.singleton old) (T.singleton new)
+{-# RULES "lazy Text replaceElem" replaceElem = replaceElemLazyText #-}
+replaceElemLazyText :: Char -> Char -> TL.Text -> TL.Text
+replaceElemLazyText old new = TL.replace (TL.singleton old) (TL.singleton new)
+
+
 -- | Monomorphic containers that can be folded.
 class MonoFoldable mono where
     -- | Map each element of a monomorphic container to a 'Monoid'
@@ -295,28 +277,38 @@ class MonoFoldable mono where
 
     -- | Map each element of a monomorphic container to an action,
     -- evaluate these actions from left to right, and ignore the results.
-    otraverse_ :: (MonoFoldable mono, Applicative f) => (Element mono -> f b) -> mono -> f ()
+    otraverse_ :: Applicative f => (Element mono -> f b) -> mono -> f ()
     otraverse_ f = ofoldr ((*>) . f) (pure ())
     {-# INLINE otraverse_ #-}
 
     -- | 'ofor_' is 'otraverse_' with its arguments flipped.
-    ofor_ :: (MonoFoldable mono, Applicative f) => mono -> (Element mono -> f b) -> f ()
+    ofor_ :: Applicative f => mono -> (Element mono -> f b) -> f ()
     ofor_ = flip otraverse_
     {-# INLINE ofor_ #-}
 
     -- | Map each element of a monomorphic container to a monadic action,
     -- evaluate these actions from left to right, and ignore the results.
-    omapM_ :: (MonoFoldable mono, Monad m) => (Element mono -> m ()) -> mono -> m ()
+#if MIN_VERSION_base(4,8,0)
+    omapM_ :: Applicative m => (Element mono -> m ()) -> mono -> m ()
+    omapM_ = otraverse_
+#else
+    omapM_ :: Monad m => (Element mono -> m ()) -> mono -> m ()
     omapM_ f = ofoldr ((>>) . f) (return ())
+#endif
     {-# INLINE omapM_ #-}
 
     -- | 'oforM_' is 'omapM_' with its arguments flipped.
-    oforM_ :: (MonoFoldable mono, Monad m) => mono -> (Element mono -> m ()) -> m ()
+#if MIN_VERSION_base(4,8,0)
+    oforM_ :: Applicative m => mono -> (Element mono -> m ()) -> m ()
     oforM_ = flip omapM_
+#else
+    oforM_ :: Monad m => mono -> (Element mono -> m ()) -> m ()
+    oforM_ = flip omapM_
+#endif
     {-# INLINE oforM_ #-}
 
     -- | Monadic fold over the elements of a monomorphic container, associating to the left.
-    ofoldlM :: (MonoFoldable mono, Monad m) => (a -> Element mono -> m a) -> a -> mono -> m a
+    ofoldlM :: Monad m => (a -> Element mono -> m a) -> a -> mono -> m a
     ofoldlM f z0 xs = ofoldr f' return xs z0
       where f' x k z = f z x >>= k
     {-# INLINE ofoldlM #-}
@@ -327,7 +319,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.ofoldMap1' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.ofoldMap1' from "Data.NonNull" for a total version of this function./
     ofoldMap1Ex :: Semigroup m => (Element mono -> m) -> mono -> m
     ofoldMap1Ex f = fromMaybe (Prelude.error "Data.MonoTraversable.ofoldMap1Ex")
                        . getOption . ofoldMap (Option . Just . f)
@@ -337,7 +329,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.ofoldr1Ex' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.ofoldr1Ex' from "Data.NonNull" for a total version of this function./
     ofoldr1Ex :: (Element mono -> Element mono -> Element mono) -> mono -> Element mono
     default ofoldr1Ex :: (t a ~ mono, a ~ Element (t a), F.Foldable t)
                            => (a -> a -> a) -> mono -> a
@@ -350,7 +342,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.ofoldl1Ex'' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.ofoldl1Ex'' from "Data.NonNull" for a total version of this function./
     ofoldl1Ex' :: (Element mono -> Element mono -> Element mono) -> mono -> Element mono
     default ofoldl1Ex' :: (t a ~ mono, a ~ Element (t a), F.Foldable t)
                             => (a -> a -> a) -> mono -> a
@@ -362,7 +354,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.head' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.head' from "Data.NonNull" for a total version of this function./
     headEx :: mono -> Element mono
     headEx = ofoldr const (Prelude.error "Data.MonoTraversable.headEx: empty")
     {-# INLINE headEx #-}
@@ -372,7 +364,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.last from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.last from "Data.NonNull" for a total version of this function./
     lastEx :: mono -> Element mono
     lastEx = ofoldl1Ex' (flip const)
     {-# INLINE lastEx #-}
@@ -393,7 +385,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.maximiumBy' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.maximiumBy' from "Data.NonNull" for a total version of this function./
     maximumByEx :: (Element mono -> Element mono -> Ordering) -> mono -> Element mono
     maximumByEx f =
         ofoldl1Ex' go
@@ -410,7 +402,7 @@ class MonoFoldable mono where
     -- Note: this is a partial function. On an empty 'MonoFoldable', it will
     -- throw an exception.
     --
-    -- /See 'Data.MinLen.minimumBy' from "Data.MinLen" for a total version of this function./
+    -- /See 'Data.NonNull.minimumBy' from "Data.NonNull" for a total version of this function./
     minimumByEx :: (Element mono -> Element mono -> Ordering) -> mono -> Element mono
     minimumByEx f =
         ofoldl1Ex' go
@@ -435,11 +427,28 @@ instance MonoFoldable S.ByteString where
         let start = Unsafe.unsafeForeignPtrToPtr fptr `plusPtr` offset
             end = start `plusPtr` len
             loop ptr
-                | ptr >= end = Unsafe.inlinePerformIO (touchForeignPtr fptr) `seq` return ()
-                | otherwise = do
-                    _ <- f (Unsafe.inlinePerformIO (peek ptr))
+                | ptr >= end = evil (touchForeignPtr fptr) `seq`
+#if MIN_VERSION_base(4,8,0)
+                    pure ()
+#else
+                    return ()
+#endif
+                | otherwise =
+#if MIN_VERSION_base(4,8,0)
+                    f (evil (peek ptr)) *>
                     loop (ptr `plusPtr` 1)
+#else
+                    f (evil (peek ptr)) >>
+                    loop (ptr `plusPtr` 1)
+#endif
         loop start
+      where
+#if MIN_VERSION_bytestring(0,10,6)
+        evil = Unsafe.accursedUnutterablePerformIO
+#else
+        evil = Unsafe.inlinePerformIO
+#endif
+        {-# INLINE evil #-}
     ofoldr1Ex = S.foldr1
     ofoldl1Ex' = S.foldl1'
     headEx = S.head
@@ -459,6 +468,8 @@ instance MonoFoldable S.ByteString where
     {-# INLINE headEx #-}
     {-# INLINE lastEx #-}
     {-# INLINE unsafeHead #-}
+{-# RULES "strict ByteString: ofoldMap = concatMap" ofoldMap = S.concatMap #-}
+
 instance MonoFoldable L.ByteString where
     ofoldMap f = ofoldr (mappend . f) mempty
     ofoldr = L.foldr
@@ -486,6 +497,8 @@ instance MonoFoldable L.ByteString where
     {-# INLINE ofoldl1Ex' #-}
     {-# INLINE headEx #-}
     {-# INLINE lastEx #-}
+{-# RULES "lazy ByteString: ofoldMap = concatMap" ofoldMap = L.concatMap #-}
+
 instance MonoFoldable T.Text where
     ofoldMap f = ofoldr (mappend . f) mempty
     ofoldr = T.foldr
@@ -511,6 +524,8 @@ instance MonoFoldable T.Text where
     {-# INLINE ofoldl1Ex' #-}
     {-# INLINE headEx #-}
     {-# INLINE lastEx #-}
+{-# RULES "strict Text: ofoldMap = concatMap" ofoldMap = T.concatMap #-}
+
 instance MonoFoldable TL.Text where
     ofoldMap f = ofoldr (mappend . f) mempty
     ofoldr = TL.foldr
@@ -535,6 +550,8 @@ instance MonoFoldable TL.Text where
     {-# INLINE ofoldl1Ex' #-}
     {-# INLINE headEx #-}
     {-# INLINE lastEx #-}
+{-# RULES "lazy Text: ofoldMap = concatMap" ofoldMap = TL.concatMap #-}
+
 instance MonoFoldable IntSet where
     ofoldMap f = ofoldr (mappend . f) mempty
     ofoldr = IntSet.foldr
@@ -561,7 +578,11 @@ instance MonoFoldable [a] where
         | i Prelude.<= 0 = GT
         | otherwise = ocompareLength xs (i - 1)
 instance MonoFoldable (Maybe a) where
+#if MIN_VERSION_base(4,8,0)
+    omapM_ _ Nothing = pure ()
+#else
     omapM_ _ Nothing = return ()
+#endif
     omapM_ f (Just x) = f x
     {-# INLINE omapM_ #-}
 instance MonoFoldable (Tree a)
@@ -610,11 +631,6 @@ instance MonoFoldable (Vector a) where
     {-# INLINE minimumByEx #-}
 instance MonoFoldable (Set e)
 instance MonoFoldable (HashSet e)
-instance MonoFoldable (DList a) where
-    otoList = DL.toList
-    headEx = DL.head
-    {-# INLINE otoList #-}
-    {-# INLINE headEx #-}
 
 instance U.Unbox a => MonoFoldable (U.Vector a) where
     ofoldMap f = ofoldr (mappend . f) mempty
@@ -700,7 +716,11 @@ instance MonoFoldable (Either a b) where
     ofoldr1Ex _ (Right x) = x
     ofoldl1Ex' _ (Left _) = Prelude.error "ofoldl1Ex' on Either"
     ofoldl1Ex' _ (Right x) = x
+#if MIN_VERSION_base(4,8,0)
+    omapM_ _ (Left _) = pure ()
+#else
     omapM_ _ (Left _) = return ()
+#endif
     omapM_ f (Right x) = f x
     {-# INLINE ofoldMap #-}
     {-# INLINE ofoldr #-}
@@ -720,7 +740,6 @@ instance F.Foldable f => MonoFoldable (ListT f a)
 instance F.Foldable f => MonoFoldable (IdentityT f a)
 instance F.Foldable f => MonoFoldable (WriterT w f a)
 instance F.Foldable f => MonoFoldable (Strict.WriterT w f a)
-instance F.Foldable f => MonoFoldable (ErrorT e f a)
 instance (F.Foldable f, F.Foldable g) => MonoFoldable (Compose f g a)
 instance (F.Foldable f, F.Foldable g) => MonoFoldable (Product f g a)
 
@@ -768,186 +787,109 @@ oor :: (Element mono ~ Bool, MonoFoldable mono) => mono -> Bool
 oor = oany id
 {-# INLINE oor #-}
 
--- | A typeclass for monomorphic containers that are 'Monoid's.
-class (MonoFoldable mono, Monoid mono) => MonoFoldableMonoid mono where -- FIXME is this really just MonoMonad?
-    -- | Map a function over a monomorphic container and combine the results.
-    oconcatMap :: (Element mono -> mono) -> mono -> mono
-    oconcatMap = ofoldMap
-    {-# INLINE oconcatMap #-}
-instance (MonoFoldable (t a), Monoid (t a)) => MonoFoldableMonoid (t a) -- FIXME
-instance MonoFoldableMonoid S.ByteString where
-    oconcatMap = S.concatMap
-    {-# INLINE oconcatMap #-}
-instance MonoFoldableMonoid L.ByteString where
-    oconcatMap = L.concatMap
-    {-# INLINE oconcatMap #-}
-instance MonoFoldableMonoid T.Text where
-    oconcatMap = T.concatMap
-    {-# INLINE oconcatMap #-}
-instance MonoFoldableMonoid TL.Text where
-    oconcatMap = TL.concatMap
-    {-# INLINE oconcatMap #-}
+-- | Synonym for 'ofoldMap'
+--
+-- @since 1.0.0
+oconcatMap :: (MonoFoldable mono, Monoid m) => (Element mono -> m) -> mono -> m
+oconcatMap = ofoldMap
 
--- | A typeclass for monomorphic containers whose elements
--- are an instance of 'Eq'.
-class (MonoFoldable mono, Eq (Element mono)) => MonoFoldableEq mono where
-    -- | Checks if the monomorphic container includes the supplied element.
-    oelem :: Element mono -> mono -> Bool
-    oelem e = List.elem e . otoList
+-- | Monoidally combine all values in the container
+--
+-- @since 1.0.0
+ofold :: (MonoFoldable mono, Monoid (Element mono)) => mono -> Element mono
+ofold = ofoldMap id
+{-# INLINE ofold #-}
 
-    -- | Checks if the monomorphic container does not include the supplied element.
-    onotElem :: Element mono -> mono -> Bool
-    onotElem e = List.notElem e . otoList
-    {-# INLINE oelem #-}
-    {-# INLINE onotElem #-}
+-- | Synonym for 'ofold'
+--
+-- @since 1.0.0
+oconcat :: (MonoFoldable mono, Monoid (Element mono)) => mono -> Element mono
+oconcat = ofold
+{-# INLINE oconcat #-}
 
-instance Eq a => MonoFoldableEq (Seq.Seq a)
-instance Eq a => MonoFoldableEq (V.Vector a)
-instance (Eq a, U.Unbox a) => MonoFoldableEq (U.Vector a)
-instance (Eq a, VS.Storable a) => MonoFoldableEq (VS.Vector a)
-instance Eq a => MonoFoldableEq (NonEmpty a)
-instance MonoFoldableEq T.Text
-instance MonoFoldableEq TL.Text
-instance MonoFoldableEq IntSet
-instance Eq a => MonoFoldableEq (Maybe a)
-instance Eq a => MonoFoldableEq (Tree a)
-instance Eq a => MonoFoldableEq (ViewL a)
-instance Eq a => MonoFoldableEq (ViewR a)
-instance Eq a => MonoFoldableEq (IntMap a)
-instance Eq a => MonoFoldableEq (Option a)
-instance Eq a => MonoFoldableEq (Identity a)
-instance Eq v => MonoFoldableEq (Map k v)
-instance Eq v => MonoFoldableEq (HashMap k v)
-instance Eq a => MonoFoldableEq (HashSet a)
-instance Eq a => MonoFoldableEq (DList a)
-instance Eq b => MonoFoldableEq (Either a b)
-instance Eq b => MonoFoldableEq (a, b)
-instance Eq a => MonoFoldableEq (Const m a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (MaybeT f a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (ListT f a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (IdentityT f a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (WriterT w f a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (Strict.WriterT w f a)
-instance (Eq a, F.Foldable f) => MonoFoldableEq (ErrorT e f a)
-instance (Eq a, F.Foldable f, F.Foldable g) => MonoFoldableEq (Compose f g a)
-instance (Eq a, F.Foldable f, F.Foldable g) => MonoFoldableEq (Product f g a)
+-- | Synonym for 'ofoldlM'
+--
+-- @since 1.0.0
+ofoldM :: (MonoFoldable mono, Monad m) => (a -> Element mono -> m a) -> a -> mono -> m a
+ofoldM = ofoldlM
+{-# INLINE ofoldM #-}
 
-instance Eq a => MonoFoldableEq [a] where
-    oelem = List.elem
-    onotElem = List.notElem
-    {-# INLINE oelem #-}
-    {-# INLINE onotElem #-}
+-- | Perform all actions in the given container
+--
+-- @since 1.0.0
+#if MIN_VERSION_base(4,8,0)
+osequence_ :: (Applicative m, MonoFoldable mono, Element mono ~ (m ())) => mono -> m ()
+#else
+osequence_ :: (Monad m, MonoFoldable mono, Element mono ~ (m ())) => mono -> m ()
+#endif
+osequence_ = omapM_ id
+{-# INLINE osequence_ #-}
 
-instance MonoFoldableEq S.ByteString where
-    oelem = S.elem
-    onotElem = S.notElem
-    {-# INLINE oelem #-}
-    {-# INLINE onotElem #-}
+-- | Checks if the monomorphic container includes the supplied element.
+oelem :: (MonoFoldable mono, Eq (Element mono)) => Element mono -> mono -> Bool
+oelem e = List.elem e . otoList
+{-# INLINE [0] oelem #-}
 
-instance MonoFoldableEq L.ByteString where
-    oelem = L.elem
-    onotElem = L.notElem
-    {-# INLINE oelem #-}
-    {-# INLINE onotElem #-}
+-- | Checks if the monomorphic container does not include the supplied element.
+onotElem :: (MonoFoldable mono, Eq (Element mono)) => Element mono -> mono -> Bool
+onotElem e = List.notElem e . otoList
+{-# INLINE [0] onotElem #-}
 
-instance (Eq a, Ord a) => MonoFoldableEq (Set a) where
-    oelem = Set.member
-    onotElem = Set.notMember
-    {-# INLINE oelem #-}
-    {-# INLINE onotElem #-}
+{-# RULES "strict ByteString elem" oelem = S.elem #-}
+{-# RULES "strict ByteString notElem" oelem = S.notElem #-}
 
+{-# RULES "lazy ByteString elem" oelem = L.elem #-}
+{-# RULES "lazy ByteString notElem" oelem = L.notElem #-}
 
--- | A typeclass for monomorphic containers whose elements
--- are an instance of 'Ord'.
-class (MonoFoldable mono, Ord (Element mono)) => MonoFoldableOrd mono where
-    -- | Get the minimum element of a monomorphic container.
-    --
-    -- Note: this is a partial function. On an empty 'MonoFoldable', it will
-    -- throw an exception.
-    --
-    -- /See 'Data.MinLen.maximum' from "Data.MinLen" for a total version of this function./
-    maximumEx :: mono -> Element mono
-    maximumEx = maximumByEx compare
-    {-# INLINE maximumEx #-}
+{-# RULES "Set elem" forall (k :: Ord k => k). oelem k = Set.member k #-}
+{-# RULES "Set notElem" forall (k :: Ord k => k). oelem k = Set.notMember k #-}
 
-    -- | Get the maximum element of a monomorphic container.
-    --
-    -- Note: this is a partial function. On an empty 'MonoFoldable', it will
-    -- throw an exception.
-    --
-    -- /See 'Data.MinLen.minimum' from "Data.MinLen" for a total version of this function./
-    minimumEx :: mono -> Element mono
-    minimumEx = minimumByEx compare
-    {-# INLINE minimumEx #-}
+-- | Get the minimum element of a monomorphic container.
+--
+-- Note: this is a partial function. On an empty 'MonoFoldable', it will
+-- throw an exception.
+--
+-- /See 'Data.NonNull.maximum' from "Data.NonNull" for a total version of this function./
+maximumEx :: (MonoFoldable mono, Ord (Element mono)) => mono -> Element mono
+maximumEx = maximumByEx compare
+{-# INLINE [0] maximumEx #-}
 
-instance MonoFoldableOrd S.ByteString where
-    maximumEx = S.maximum
-    {-# INLINE maximumEx #-}
-    minimumEx = S.minimum
-    {-# INLINE minimumEx #-}
-instance MonoFoldableOrd L.ByteString where
-    maximumEx = L.maximum
-    {-# INLINE maximumEx #-}
-    minimumEx = L.minimum
-    {-# INLINE minimumEx #-}
-instance MonoFoldableOrd T.Text where
-    maximumEx = T.maximum
-    {-# INLINE maximumEx #-}
-    minimumEx = T.minimum
-    {-# INLINE minimumEx #-}
-instance MonoFoldableOrd TL.Text where
-    maximumEx = TL.maximum
-    {-# INLINE maximumEx #-}
-    minimumEx = TL.minimum
-    {-# INLINE minimumEx #-}
-instance MonoFoldableOrd IntSet
-instance Ord a => MonoFoldableOrd [a]
-instance Ord a => MonoFoldableOrd (Maybe a)
-instance Ord a => MonoFoldableOrd (Tree a)
-instance Ord a => MonoFoldableOrd (Seq a)
-instance Ord a => MonoFoldableOrd (ViewL a)
-instance Ord a => MonoFoldableOrd (ViewR a)
-instance Ord a => MonoFoldableOrd (IntMap a)
-instance Ord a => MonoFoldableOrd (Option a)
-instance Ord a => MonoFoldableOrd (NonEmpty a)
-instance Ord a => MonoFoldableOrd (Identity a)
-instance Ord v => MonoFoldableOrd (Map k v)
-instance Ord v => MonoFoldableOrd (HashMap k v)
-instance Ord a => MonoFoldableOrd (Vector a) where
-    maximumEx   = V.maximum
-    minimumEx   = V.minimum
-    {-# INLINE maximumEx #-}
-    {-# INLINE minimumEx #-}
-instance Ord e => MonoFoldableOrd (Set e)
-instance Ord e => MonoFoldableOrd (HashSet e)
-instance (U.Unbox a, Ord a) => MonoFoldableOrd (U.Vector a) where
-    maximumEx   = U.maximum
-    minimumEx   = U.minimum
-    {-# INLINE maximumEx #-}
-    {-# INLINE minimumEx #-}
-instance (Ord a, VS.Storable a) => MonoFoldableOrd (VS.Vector a) where
-    maximumEx   = VS.maximum
-    minimumEx   = VS.minimum
-    {-# INLINE maximumEx #-}
-    {-# INLINE minimumEx #-}
-instance Ord b => MonoFoldableOrd (Either a b) where
-instance Ord a => MonoFoldableOrd (DList a)
-instance Ord b => MonoFoldableOrd (a, b)
-instance Ord a => MonoFoldableOrd (Const m a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (MaybeT f a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (ListT f a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (IdentityT f a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (WriterT w f a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (Strict.WriterT w f a)
-instance (Ord a, F.Foldable f) => MonoFoldableOrd (ErrorT e f a)
-instance (Ord a, F.Foldable f, F.Foldable g) => MonoFoldableOrd (Compose f g a)
-instance (Ord a, F.Foldable f, F.Foldable g) => MonoFoldableOrd (Product f g a)
+-- | Get the maximum element of a monomorphic container.
+--
+-- Note: this is a partial function. On an empty 'MonoFoldable', it will
+-- throw an exception.
+--
+-- /See 'Data.NonNull.minimum' from "Data.NonNull" for a total version of this function./
+minimumEx :: (MonoFoldable mono, Ord (Element mono)) => mono -> Element mono
+minimumEx = minimumByEx compare
+{-# INLINE [0] minimumEx #-}
+
+{-# RULES "strict ByteString maximumEx" maximumEx = S.maximum #-}
+{-# RULES "strict ByteString minimumEx" minimumEx = S.minimum #-}
+
+{-# RULES "lazy ByteString maximumEx" maximumEx = L.maximum #-}
+{-# RULES "lazy ByteString minimumEx" minimumEx = L.minimum #-}
+
+{-# RULES "strict Text maximumEx" maximumEx = T.maximum #-}
+{-# RULES "strict Text minimumEx" minimumEx = T.minimum #-}
+
+{-# RULES "lazy Text maximumEx" maximumEx = TL.maximum #-}
+{-# RULES "lazy Text minimumEx" minimumEx = TL.minimum #-}
+
+{-# RULES "boxed Vector maximumEx" maximumEx = V.maximum #-}
+{-# RULES "boxed Vector minimumEx" minimumEx = V.minimum #-}
+
+{-# RULES "unboxed Vector maximumEx" forall (u :: U.Unbox a => U.Vector a). maximumEx u = U.maximum u #-}
+{-# RULES "unboxed Vector minimumEx" forall (u :: U.Unbox a => U.Vector a). minimumEx u = U.minimum u #-}
+
+{-# RULES "storable Vector maximumEx" forall (v :: VS.Storable a => VS.Vector a). maximumEx v = VS.maximum v #-}
+{-# RULES "storable Vector minimumEx" forall (v :: VS.Storable a => VS.Vector a). minimumEx v = VS.minimum v #-}
 
 -- | Safe version of 'maximumEx'.
 --
 -- Returns 'Nothing' instead of throwing an exception when
 -- encountering an empty monomorphic container.
-maximumMay :: MonoFoldableOrd mono => mono -> Maybe (Element mono)
+maximumMay :: (MonoFoldable mono, Ord (Element mono)) => mono -> Maybe (Element mono)
 maximumMay mono
     | onull mono = Nothing
     | otherwise = Just (maximumEx mono)
@@ -970,7 +912,7 @@ maximumByMay f mono
 --
 -- Returns 'Nothing' instead of throwing an exception when
 -- encountering an empty monomorphic container.
-minimumMay :: MonoFoldableOrd mono => mono -> Maybe (Element mono)
+minimumMay :: (MonoFoldable mono, Ord (Element mono)) => mono -> Maybe (Element mono)
 minimumMay mono
     | onull mono = Nothing
     | otherwise = Just (minimumEx mono)
@@ -1001,32 +943,45 @@ class (MonoFunctor mono, MonoFoldable mono) => MonoTraversable mono where
     -- | Map each element of a monomorphic container to a monadic action,
     -- evaluate these actions from left to right, and
     -- collect the results.
+#if MIN_VERSION_base(4,8,0)
+    omapM :: Applicative m => (Element mono -> m (Element mono)) -> mono -> m mono
+    omapM = otraverse
+#else
     omapM :: Monad m => (Element mono -> m (Element mono)) -> mono -> m mono
     default omapM :: (Traversable t, mono ~ t a, a ~ Element mono, Monad m) => (Element mono -> m (Element mono)) -> mono -> m mono
     omapM = mapM
+#endif
     {-# INLINE otraverse #-}
     {-# INLINE omapM #-}
 
 instance MonoTraversable S.ByteString where
     otraverse f = fmap S.pack . traverse f . S.unpack
-    omapM f = liftM S.pack . mapM f . S.unpack
     {-# INLINE otraverse #-}
+#if !MIN_VERSION_base(4,8,0)
+    omapM f = liftM S.pack . mapM f . S.unpack
     {-# INLINE omapM #-}
+#endif
 instance MonoTraversable L.ByteString where
     otraverse f = fmap L.pack . traverse f . L.unpack
-    omapM f = liftM L.pack . mapM f . L.unpack
     {-# INLINE otraverse #-}
+#if !MIN_VERSION_base(4,8,0)
+    omapM f = liftM L.pack . mapM f . L.unpack
     {-# INLINE omapM #-}
+#endif
 instance MonoTraversable T.Text where
     otraverse f = fmap T.pack . traverse f . T.unpack
-    omapM f = liftM T.pack . mapM f . T.unpack
     {-# INLINE otraverse #-}
+#if !MIN_VERSION_base(4,8,0)
+    omapM f = liftM T.pack . mapM f . T.unpack
     {-# INLINE omapM #-}
+#endif
 instance MonoTraversable TL.Text where
     otraverse f = fmap TL.pack . traverse f . TL.unpack
-    omapM f = liftM TL.pack . mapM f . TL.unpack
     {-# INLINE otraverse #-}
+#if !MIN_VERSION_base(4,8,0)
+    omapM f = liftM TL.pack . mapM f . TL.unpack
     {-# INLINE omapM #-}
+#endif
 instance MonoTraversable [a]
 instance MonoTraversable (Maybe a)
 instance MonoTraversable (Tree a)
@@ -1036,28 +991,40 @@ instance MonoTraversable (ViewR a)
 instance MonoTraversable (IntMap a)
 instance MonoTraversable (Option a)
 instance MonoTraversable (NonEmpty a)
-instance MonoTraversable (DList a) where
-     otraverse f = fmap DL.fromList . traverse f . DL.toList
-     omapM f = liftM DL.fromList . mapM f . DL.toList
 instance MonoTraversable (Identity a)
 instance MonoTraversable (Map k v)
 instance MonoTraversable (HashMap k v)
 instance MonoTraversable (Vector a)
 instance U.Unbox a => MonoTraversable (U.Vector a) where
+    -- FIXME do something more efficient
     otraverse f = fmap U.fromList . traverse f . U.toList
+#if MIN_VERSION_base(4,8,0)
+    omapM = otraverse
+#else
     omapM = U.mapM
+#endif
     {-# INLINE otraverse #-}
     {-# INLINE omapM #-}
 instance VS.Storable a => MonoTraversable (VS.Vector a) where
+    -- FIXME do something more efficient
     otraverse f = fmap VS.fromList . traverse f . VS.toList
+#if MIN_VERSION_base(4,8,0)
+    omapM = otraverse
+#else
     omapM = VS.mapM
+#endif
     {-# INLINE otraverse #-}
     {-# INLINE omapM #-}
 instance MonoTraversable (Either a b) where
     otraverse _ (Left a) = pure (Left a)
     otraverse f (Right b) = fmap Right (f b)
+#if MIN_VERSION_base(4,8,0)
+    omapM _ (Left a) = pure (Left a)
+    omapM f (Right b) = fmap Right (f b)
+#else
     omapM _ (Left a) = return (Left a)
     omapM f (Right b) = liftM Right (f b)
+#endif
     {-# INLINE otraverse #-}
     {-# INLINE omapM #-}
 instance MonoTraversable (a, b)
@@ -1067,7 +1034,6 @@ instance Traversable f => MonoTraversable (ListT f a)
 instance Traversable f => MonoTraversable (IdentityT f a)
 instance Traversable f => MonoTraversable (WriterT w f a)
 instance Traversable f => MonoTraversable (Strict.WriterT w f a)
-instance Traversable f => MonoTraversable (ErrorT e f a)
 instance (Traversable f, Traversable g) => MonoTraversable (Compose f g a)
 instance (Traversable f, Traversable g) => MonoTraversable (Product f g a)
 
@@ -1077,7 +1043,11 @@ ofor = flip otraverse
 {-# INLINE ofor #-}
 
 -- | 'oforM' is 'omapM' with its arguments flipped.
+#if MIN_VERSION_base(4,8,0)
+oforM :: (MonoTraversable mono, Applicative f) => mono -> (Element mono -> f (Element mono)) -> f mono
+#else
 oforM :: (MonoTraversable mono, Monad f) => mono -> (Element mono -> f (Element mono)) -> f mono
+#endif
 oforM = flip omapM
 {-# INLINE oforM #-}
 
@@ -1144,7 +1114,6 @@ instance MonoPointed (Option a)
 instance MonoPointed (NonEmpty a)
 instance MonoPointed (Identity a)
 instance MonoPointed (Vector a)
-instance MonoPointed (DList a)
 instance MonoPointed (IO a)
 instance MonoPointed (ZipList a)
 instance MonoPointed (r -> a)
@@ -1153,7 +1122,6 @@ instance Monoid m => MonoPointed (Const m a)
 instance Monad m => MonoPointed (WrappedMonad m a)
 instance Applicative m => MonoPointed (ListT m a)
 instance Applicative m => MonoPointed (IdentityT m a)
-instance Applicative f => MonoPointed (WrappedApplicative f a)
 instance Arrow a => MonoPointed (WrappedArrow a b c)
 instance (Monoid w, Applicative m) => MonoPointed (WriterT w m a)
 instance (Monoid w, Applicative m) => MonoPointed (Strict.WriterT w m a)
@@ -1161,8 +1129,6 @@ instance Applicative m => MonoPointed (ReaderT r m a)
 instance MonoPointed (ContT r m a)
 instance (Applicative f, Applicative g) => MonoPointed (Compose f g a)
 instance (Applicative f, Applicative g) => MonoPointed (Product f g a)
-instance MonoPointed (Cokleisli w a b)
-instance Applicative f => MonoPointed (Static f a b)
 
 -- Not Applicative
 instance MonoPointed (Seq a) where
@@ -1186,12 +1152,6 @@ instance MonoPointed (Set a) where
 instance Hashable a => MonoPointed (HashSet a) where
     opoint = HashSet.singleton
     {-# INLINE opoint #-}
-instance Applicative m => MonoPointed (ErrorT e m a) where
-    opoint = ErrorT . pure . Right
-    {-# INLINE opoint #-}
-instance MonoPointed (MaybeApply f a) where
-    opoint = MaybeApply . Right
-    {-# INLINE opoint #-}
 instance Applicative f => MonoPointed (MaybeT f a) where
     opoint = MaybeT . fmap Just . pure
     {-# INLINE opoint #-}
@@ -1243,30 +1203,6 @@ class MonoFunctor mono => MonoComonad mono where
     -- glimpsed from the given function.
     oextend :: (mono -> Element mono) -> mono -> mono
 
-    default oextract :: (Comonad w, (w a) ~ mono, Element (w a) ~ a)
-                   => mono -> Element mono
-    oextract = extract
-    {-# INLINE oextract #-}
-    default oextend :: (Comonad w, (w a) ~ mono, Element (w a) ~ a)
-                   => (mono -> Element mono) -> mono -> mono
-    oextend = extend
-    {-# INLINE oextend #-}
-
--- Comonad
-instance MonoComonad (Tree a)
-instance MonoComonad (NonEmpty a)
-instance MonoComonad (Identity a)
-instance Monoid m => MonoComonad (m -> a)
-instance MonoComonad (e, a)
-instance MonoComonad (Arg a b)
-instance Comonad w => MonoComonad (IdentityT w a)
-instance Comonad w => MonoComonad (EnvT e w a)
-instance Comonad w => MonoComonad (StoreT s w a)
-instance (Comonad w, Monoid m) => MonoComonad (TracedT m w a)
-#if !MIN_VERSION_comonad(5,0,0)
-instance (Comonad f, Comonad g) => MonoComonad (Coproduct f g a)
-#endif
-
 -- Not Comonad
 instance MonoComonad (ViewL a) where
     oextract ~(x :< _) = x
@@ -1288,3 +1224,57 @@ instance MonoComonad (ViewR a) where
                        EmptyR  -> Seq.empty
                        ys :> y -> ys Seq.|> y
         ) :> f w
+
+-- | Containers which, when two values are combined, the combined length is no
+-- less than the larger of the two inputs. In code:
+--
+-- @
+-- olength (x <> y) >= max (olength x) (olength y)
+-- @
+--
+-- This class has no methods, and is simply used to assert that this law holds,
+-- in order to provide guarantees of correctness (see, for instance,
+-- "Data.NonNull").
+--
+-- This should have a @Semigroup@ superclass constraint, however, due to
+-- @Semigroup@ only recently moving to base, some packages do not provide
+-- instances.
+class MonoFoldable mono => GrowingAppend mono
+
+instance GrowingAppend (Seq.Seq a)
+instance GrowingAppend [a]
+instance GrowingAppend (V.Vector a)
+instance U.Unbox a => GrowingAppend (U.Vector a)
+instance VS.Storable a => GrowingAppend (VS.Vector a)
+instance GrowingAppend S.ByteString
+instance GrowingAppend L.ByteString
+instance GrowingAppend T.Text
+instance GrowingAppend TL.Text
+instance GrowingAppend (NonEmpty a)
+instance Ord k => GrowingAppend (Map k v)
+instance (Eq k, Hashable k) => GrowingAppend (HashMap k v)
+instance Ord v => GrowingAppend (Set.Set v)
+instance (Eq v, Hashable v) => GrowingAppend (HashSet.HashSet v)
+instance GrowingAppend IntSet.IntSet
+instance GrowingAppend (IntMap v)
+
+-- | 'intercalate' @seq seqs@ inserts @seq@ in between @seqs@ and
+-- concatenates the result.
+--
+-- @since 1.0.0
+ointercalate :: (MonoFoldable mono, Monoid (Element mono))
+             => Element mono
+             -> mono
+             -> Element mono
+ointercalate x = mconcat . List.intersperse x . otoList
+{-# INLINE [0] ointercalate #-}
+{-# RULES "ointercalate list" forall x.
+        ointercalate x = List.intercalate x . otoList #-}
+{-# RULES "intercalate ByteString" forall x.
+        ointercalate x = S.intercalate x . otoList #-}
+{-# RULES "intercalate LByteString" forall x.
+        ointercalate x = L.intercalate x . otoList #-}
+{-# RULES "intercalate Text" forall x.
+        ointercalate x = T.intercalate x . otoList #-}
+{-# RULES "intercalate LText" forall x.
+        ointercalate x = TL.intercalate x . otoList #-}
