@@ -3,11 +3,13 @@
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE ConstraintKinds #-}
--- | Warning: This module should be considered highly experimental.
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE FunctionalDependencies #-}
+-- | Abstractions over sequential data structures, like lists and vectors.
 module Data.Sequences where
 
 import Data.Maybe (fromJust, isJust)
-import Data.Monoid (Monoid, mappend, mconcat, mempty)
+import Data.Monoid (Monoid, mconcat, mempty)
 import Data.MonoTraversable
 import Data.Int (Int64, Int)
 import qualified Data.List as List
@@ -23,23 +25,24 @@ import Control.Category
 import Control.Arrow ((***), first, second)
 import Control.Monad (liftM)
 import qualified Data.Sequence as Seq
-import qualified Data.DList as DList
 import qualified Data.Vector as V
 import qualified Data.Vector.Unboxed as U
 import qualified Data.Vector.Storable as VS
 import Data.String (IsString)
 import qualified Data.List.NonEmpty as NE
 import qualified Data.ByteString.Unsafe as SU
-import Data.GrowingAppend
-import Data.Vector.Instances ()
 import qualified Data.Vector.Generic as VG
 import qualified Data.Vector.Algorithms.Merge as VAM
 import Data.Ord (comparing)
+import qualified Data.Text.Encoding as T
+import qualified Data.Text.Lazy.Encoding as TL
+import Data.Text.Encoding.Error (lenientDecode)
+import Data.Word (Word8)
 
--- | 'SemiSequence' was created to share code between 'IsSequence' and 'MinLen'.
+-- | 'SemiSequence' was created to share code between 'IsSequence' and 'NonNull'.
 --
 -- @Semi@ means 'SemiGroup'
--- A 'SemiSequence' can accomodate a 'SemiGroup' such as 'NonEmpty' or 'MinLen'
+-- A 'SemiSequence' can accomodate a 'SemiGroup' such as 'NonEmpty' or 'NonNull'
 -- A Monoid should be able to fill out 'IsSequence'.
 --
 -- 'SemiSequence' operations maintain the same type because they all maintain the same number of elements or increase them.
@@ -48,6 +51,10 @@ import Data.Ord (comparing)
 -- This type-changing function exists on 'NonNull' as 'nfilter'
 --
 -- 'filter' and other such functions are placed in 'IsSequence'
+--
+-- /NOTE/: Like 'GrowingAppend', ideally we'd have a @Semigroup@ superclass
+-- constraint here, but that would pull in more dependencies to this package
+-- than desired.
 class (Integral (Index seq), GrowingAppend seq) => SemiSequence seq where
     -- | The type of the index of a sequence.
     type Index seq
@@ -368,6 +375,18 @@ class (Monoid seq, MonoTraversable seq, SemiSequence seq, MonoPointed seq) => Is
     tailEx :: seq -> seq
     tailEx = snd . maybe (error "Data.Sequences.tailEx") id . uncons
 
+    -- | Safe version of 'tailEx'.
+    --
+    -- Returns 'Nothing' instead of throwing an exception when encountering
+    -- an empty monomorphic container.
+    --
+    -- @since 1.0.0
+    tailMay :: seq -> Maybe seq
+    tailMay seq
+        | onull seq = Nothing
+        | otherwise = Just (tailEx seq)
+    {-# INLINE tailMay #-}
+
     -- | __Unsafe__
     --
     -- Get the init of a sequence, throw an exception if the sequence is empty.
@@ -379,6 +398,18 @@ class (Monoid seq, MonoTraversable seq, SemiSequence seq, MonoPointed seq) => Is
     initEx :: seq -> seq
     initEx = fst . maybe (error "Data.Sequences.initEx") id . unsnoc
 
+    -- | Safe version of 'initEx'.
+    --
+    -- Returns 'Nothing' instead of throwing an exception when encountering
+    -- an empty monomorphic container.
+    --
+    -- @since 1.0.0
+    initMay :: IsSequence seq => seq -> Maybe seq
+    initMay seq
+        | onull seq = Nothing
+        | otherwise = Just (initEx seq)
+    {-# INLINE initMay #-}
+
     -- | Equivalent to 'tailEx'.
     unsafeTail :: seq -> seq
     unsafeTail = tailEx
@@ -410,13 +441,6 @@ class (Monoid seq, MonoTraversable seq, SemiSequence seq, MonoPointed seq) => Is
     unsafeIndex :: seq -> Index seq -> Element seq
     unsafeIndex = indexEx
 
-    -- | 'intercalate' @seq seqs@ inserts @seq@ in between @seqs@ and
-    -- concatenates the result.
-    --
-    -- Since 0.9.3
-    intercalate :: seq -> [seq] -> seq
-    intercalate = defaultIntercalate
-
     -- | 'splitWhen' splits a sequence into components delimited by separators,
     -- where the predicate returns True for a separator element. The resulting
     -- components do not contain the separators. Two adjacent separators result
@@ -478,12 +502,6 @@ defaultSortBy :: IsSequence seq => (Element seq -> Element seq -> Ordering) -> s
 defaultSortBy f = fromList . sortBy f . otoList
 {-# INLINE defaultSortBy #-}
 
--- | Default 'intercalate'
-defaultIntercalate :: (IsSequence seq) => seq -> [seq] -> seq
-defaultIntercalate _ [] = mempty
-defaultIntercalate s (seq:seqs) = mconcat (seq : List.map (s `mappend`) seqs)
-{-# INLINE defaultIntercalate #-}
-
 -- | Use 'splitWhen' from "Data.List.Split"
 defaultSplitWhen :: IsSequence seq => (Element seq -> Bool) -> seq -> [seq]
 defaultSplitWhen f = List.map fromList . List.splitWhen f . otoList
@@ -566,7 +584,6 @@ instance IsSequence [a] where
       where
         (matches, nonMatches) = partition ((== f head) . f) tail
     groupAllOn _ [] = []
-    intercalate = List.intercalate
     splitWhen = List.splitWhen
     {-# INLINE fromList #-}
     {-# INLINE break #-}
@@ -585,7 +602,6 @@ instance IsSequence [a] where
     {-# INLINE replicateM #-}
     {-# INLINE groupBy #-}
     {-# INLINE groupAllOn #-}
-    {-# INLINE intercalate #-}
     {-# INLINE splitWhen #-}
 
 instance SemiSequence (NE.NonEmpty a) where
@@ -643,7 +659,6 @@ instance IsSequence S.ByteString where
     unsafeTail = SU.unsafeTail
     splitWhen f s | S.null s = [S.empty]
                   | otherwise = S.splitWith f s
-    intercalate = S.intercalate
     {-# INLINE fromList #-}
     {-# INLINE break #-}
     {-# INLINE span #-}
@@ -664,7 +679,6 @@ instance IsSequence S.ByteString where
     {-# INLINE initEx #-}
     {-# INLINE unsafeTail #-}
     {-# INLINE splitWhen #-}
-    {-# INLINE intercalate #-}
 
     index bs i
         | i >= S.length bs = Nothing
@@ -710,7 +724,6 @@ instance IsSequence T.Text where
     tailEx = T.tail
     initEx = T.init
     splitWhen = T.split
-    intercalate = T.intercalate
     {-# INLINE fromList #-}
     {-# INLINE break #-}
     {-# INLINE span #-}
@@ -728,7 +741,6 @@ instance IsSequence T.Text where
     {-# INLINE tailEx #-}
     {-# INLINE initEx #-}
     {-# INLINE splitWhen #-}
-    {-# INLINE intercalate #-}
 
     index t i
         | i >= T.length t = Nothing
@@ -775,7 +787,6 @@ instance IsSequence L.ByteString where
     initEx = L.init
     splitWhen f s | L.null s = [L.empty]
                   | otherwise = L.splitWith f s
-    intercalate = L.intercalate
     {-# INLINE fromList #-}
     {-# INLINE break #-}
     {-# INLINE span #-}
@@ -793,7 +804,6 @@ instance IsSequence L.ByteString where
     {-# INLINE tailEx #-}
     {-# INLINE initEx #-}
     {-# INLINE splitWhen #-}
-    {-# INLINE intercalate #-}
 
     indexEx = L.index
     unsafeIndex = L.index
@@ -835,7 +845,6 @@ instance IsSequence TL.Text where
     tailEx = TL.tail
     initEx = TL.init
     splitWhen = TL.split
-    intercalate = TL.intercalate
     {-# INLINE fromList #-}
     {-# INLINE break #-}
     {-# INLINE span #-}
@@ -853,7 +862,6 @@ instance IsSequence TL.Text where
     {-# INLINE tailEx #-}
     {-# INLINE initEx #-}
     {-# INLINE splitWhen #-}
-    {-# INLINE intercalate #-}
 
     indexEx = TL.index
     unsafeIndex = TL.index
@@ -927,30 +935,6 @@ instance IsSequence (Seq.Seq a) where
     {-# INLINE indexEx #-}
     {-# INLINE unsafeIndex #-}
 
-instance SemiSequence (DList.DList a) where
-    type Index (DList.DList a) = Int
-    cons = DList.cons
-    snoc = DList.snoc
-
-    reverse = defaultReverse
-    sortBy = defaultSortBy
-    intersperse = defaultIntersperse
-    find = defaultFind
-    {-# INLINE intersperse #-}
-    {-# INLINE reverse #-}
-    {-# INLINE find #-}
-    {-# INLINE sortBy #-}
-    {-# INLINE cons #-}
-    {-# INLINE snoc #-}
-
-instance IsSequence (DList.DList a) where
-    fromList = DList.fromList
-    replicate = DList.replicate
-    tailEx = DList.tail
-    {-# INLINE fromList #-}
-    {-# INLINE replicate #-}
-    {-# INLINE tailEx #-}
-
 instance SemiSequence (V.Vector a) where
     type Index (V.Vector a) = Int
     reverse = V.reverse
@@ -1173,268 +1157,236 @@ instance VS.Storable a => IsSequence (VS.Vector a) where
     {-# INLINE indexEx #-}
     {-# INLINE unsafeIndex #-}
 
--- | A typeclass for sequences whose elements have the 'Eq' typeclass
-class (MonoFoldableEq seq, IsSequence seq, Eq (Element seq)) => EqSequence seq where
+-- | @'splitElem'@ splits a sequence into components delimited by separator
+-- element. It's equivalent to 'splitWhen' with equality predicate:
+--
+-- > splitElem sep === splitWhen (== sep)
+--
+-- Since 0.9.3
+splitElem :: (IsSequence seq, Eq (Element seq)) => Element seq -> seq -> [seq]
+splitElem x = splitWhen (== x)
 
-    -- | @'splitElem'@ splits a sequence into components delimited by separator
-    -- element. It's equivalent to 'splitWhen' with equality predicate:
-    --
-    -- > splitElem sep === splitWhen (== sep)
-    --
-    -- Since 0.9.3
-    splitElem :: Element seq -> seq -> [seq]
-    splitElem x = splitWhen (== x)
+-- | @'splitSeq'@ splits a sequence into components delimited by
+-- separator subsequence. 'splitSeq' is the right inverse of 'intercalate':
+--
+-- > ointercalate x . splitSeq x === id
+--
+-- 'splitElem' can be considered a special case of 'splitSeq'
+--
+-- > splitSeq (singleton sep) === splitElem sep
+--
+-- @'splitSeq' mempty@ is another special case: it splits just before each
+-- element, and in line with 'splitWhen' rules, it has at least one output
+-- component:
+--
+-- @
+-- > 'splitSeq' "" ""
+-- [""]
+-- > 'splitSeq' "" "a"
+-- ["", "a"]
+-- > 'splitSeq' "" "ab"
+-- ["", "a", "b"]
+-- @
+--
+-- Since 0.9.3
+splitSeq :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> [seq]
+splitSeq sep = List.map fromList . List.splitOn (otoList sep) . otoList
 
-    -- | @'splitSeq'@ splits a sequence into components delimited by
-    -- separator subsequence. 'splitSeq' is the right inverse of 'intercalate':
-    --
-    -- > intercalate x . splitSeq x === id
-    --
-    -- 'splitElem' can be considered a special case of 'splitSeq'
-    --
-    -- > splitSeq (singleton sep) === splitElem sep
-    --
-    -- @'splitSeq' mempty@ is another special case: it splits just before each
-    -- element, and in line with 'splitWhen' rules, it has at least one output
-    -- component:
-    --
-    -- @
-    -- > 'splitSeq' "" ""
-    -- [""]
-    -- > 'splitSeq' "" "a"
-    -- ["", "a"]
-    -- > 'splitSeq' "" "ab"
-    -- ["", "a", "b"]
-    -- @
-    --
-    -- Since 0.9.3
-    splitSeq :: seq -> seq -> [seq]
-    splitSeq = defaultSplitOn
+-- | @'replaceSeq' old new@ replaces all @old@ subsequences with @new@.
+--
+-- > replaceSeq old new === ointercalate new . splitSeq old
+--
+-- @since 1.0.1
 
-    -- | 'stripPrefix' drops the given prefix from a sequence.
-    -- It returns 'Nothing' if the sequence did not start with the prefix
-    -- given, or 'Just' the sequence after the prefix, if it does.
-    --
-    -- @
-    -- > 'stripPrefix' "foo" "foobar"
-    -- 'Just' "foo"
-    -- > 'stripPrefix' "abc" "foobar"
-    -- 'Nothing'
-    -- @
-    stripPrefix :: seq -> seq -> Maybe seq
-    stripPrefix x y = fmap fromList (otoList x `stripPrefix` otoList y)
+replaceSeq :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> seq -> seq
+replaceSeq old new = ointercalate new . splitSeq old
 
-    -- | 'stripSuffix' drops the given suffix from a sequence.
-    -- It returns 'Nothing' if the sequence did not end with the suffix
-    -- given, or 'Just' the sequence before the suffix, if it does.
-    --
-    -- @
-    -- > 'stripSuffix' "bar" "foobar"
-    -- 'Just' "foo"
-    -- > 'stripSuffix' "abc" "foobar"
-    -- 'Nothing'
-    -- @
-    stripSuffix :: seq -> seq -> Maybe seq
-    stripSuffix x y = fmap fromList (otoList x `stripSuffix` otoList y)
+-- | 'stripPrefix' drops the given prefix from a sequence.
+-- It returns 'Nothing' if the sequence did not start with the prefix
+-- given, or 'Just' the sequence after the prefix, if it does.
+--
+-- @
+-- > 'stripPrefix' "foo" "foobar"
+-- 'Just' "foo"
+-- > 'stripPrefix' "abc" "foobar"
+-- 'Nothing'
+-- @
+stripPrefix :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> Maybe seq
+stripPrefix x y = fmap fromList (otoList x `List.stripPrefix` otoList y)
 
-    -- | 'isPrefixOf' takes two sequences and returns 'True' if the first
-    -- sequence is a prefix of the second.
-    isPrefixOf :: seq -> seq -> Bool
-    isPrefixOf x y = otoList x `isPrefixOf` otoList y
+-- | 'stripSuffix' drops the given suffix from a sequence.
+-- It returns 'Nothing' if the sequence did not end with the suffix
+-- given, or 'Just' the sequence before the suffix, if it does.
+--
+-- @
+-- > 'stripSuffix' "bar" "foobar"
+-- 'Just' "foo"
+-- > 'stripSuffix' "abc" "foobar"
+-- 'Nothing'
+-- @
+stripSuffix :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> Maybe seq
+stripSuffix x y =
+    fmap fromList (otoList x `stripSuffixList` otoList y)
+  where
+    stripSuffixList :: Eq a => [a] -> [a] -> Maybe [a]
+    stripSuffixList x' y' = fmap reverse (stripPrefix (reverse x') (reverse y'))
 
-    -- | 'isSuffixOf' takes two sequences and returns 'True' if the first
-    -- sequence is a suffix of the second.
-    isSuffixOf :: seq -> seq -> Bool
-    isSuffixOf x y = otoList x `isSuffixOf` otoList y
+-- | 'isPrefixOf' takes two sequences and returns 'True' if the first
+-- sequence is a prefix of the second.
+isPrefixOf :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> Bool
+isPrefixOf x y = otoList x `List.isPrefixOf` otoList y
 
-    -- | 'isInfixOf' takes two sequences and returns 'true' if the first
-    -- sequence is contained, wholly and intact, anywhere within the second.
-    isInfixOf :: seq -> seq -> Bool
-    isInfixOf x y = otoList x `isInfixOf` otoList y
+-- | 'isSuffixOf' takes two sequences and returns 'True' if the first
+-- sequence is a suffix of the second.
+isSuffixOf :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> Bool
+isSuffixOf x y = otoList x `List.isSuffixOf` otoList y
 
-    -- | Equivalent to @'groupBy' (==)@
-    group :: seq -> [seq]
-    group = groupBy (==)
+-- | 'isInfixOf' takes two sequences and returns 'true' if the first
+-- sequence is contained, wholly and intact, anywhere within the second.
+isInfixOf :: (IsSequence seq, Eq (Element seq)) => seq -> seq -> Bool
+isInfixOf x y = otoList x `List.isInfixOf` otoList y
 
-    -- | Similar to standard 'group', but operates on the whole collection,
-    -- not just the consecutive items.
-    --
-    -- Equivalent to @'groupAllOn' id@
-    groupAll :: seq -> [seq]
-    groupAll = groupAllOn id
+-- | Equivalent to @'groupBy' (==)@
+group :: (IsSequence seq, Eq (Element seq)) => seq -> [seq]
+group = groupBy (==)
 
-    -- |
-    --
-    -- @since 0.10.2
-    delete :: Element seq -> seq -> seq
-    delete = deleteBy (==)
+-- | Similar to standard 'group', but operates on the whole collection,
+-- not just the consecutive items.
+--
+-- Equivalent to @'groupAllOn' id@
+groupAll :: (IsSequence seq, Eq (Element seq)) => seq -> [seq]
+groupAll = groupAllOn id
 
-    -- |
-    --
-    -- @since 0.10.2
-    deleteBy :: (Element seq -> Element seq -> Bool) -> Element seq -> seq -> seq
-    deleteBy eq x = fromList . List.deleteBy eq x . otoList
-    {-# INLINE splitElem #-}
-    {-# INLINE splitSeq #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE groupAll #-}
-    {-# INLINE delete #-}
-    {-# INLINE deleteBy #-}
+-- |
+--
+-- @since 0.10.2
+delete :: (IsSequence seq, Eq (Element seq)) => Element seq -> seq -> seq
+delete = deleteBy (==)
 
-{-# DEPRECATED elem "use oelem" #-}
-elem :: EqSequence seq => Element seq -> seq -> Bool
-elem = oelem
+-- |
+--
+-- @since 0.10.2
+deleteBy :: (IsSequence seq, Eq (Element seq)) => (Element seq -> Element seq -> Bool) -> Element seq -> seq -> seq
+deleteBy eq x = fromList . List.deleteBy eq x . otoList
 
-{-# DEPRECATED notElem "use onotElem" #-}
-notElem :: EqSequence seq => Element seq -> seq -> Bool
-notElem = onotElem
+{-# INLINE [0] splitElem #-}
+{-# INLINE [0] splitSeq #-}
+{-# INLINE [0] replaceSeq #-}
+{-# INLINE [0] isPrefixOf #-}
+{-# INLINE [0] isSuffixOf #-}
+{-# INLINE [0] isInfixOf #-}
+{-# INLINE [0] stripPrefix #-}
+{-# INLINE [0] stripSuffix #-}
+{-# INLINE [0] group #-}
+{-# INLINE [0] groupAll #-}
+{-# INLINE [0] delete #-}
+{-# INLINE [0] deleteBy #-}
 
--- | Use 'splitOn' from "Data.List.Split"
-defaultSplitOn :: EqSequence s => s -> s -> [s]
-defaultSplitOn sep = List.map fromList . List.splitOn (otoList sep) . otoList
+{-# RULES "list splitSeq" splitSeq = List.splitOn #-}
+{-# RULES "list stripPrefix" stripPrefix = List.stripPrefix #-}
+{-# RULES "list isPrefixOf" isPrefixOf = List.isPrefixOf #-}
+{-# RULES "list isSuffixOf" isSuffixOf = List.isSuffixOf #-}
+{-# RULES "list isInfixOf" isInfixOf = List.isInfixOf #-}
+{-# RULES "list delete" delete = List.delete #-}
+{-# RULES "list deleteBy" deleteBy = List.deleteBy #-}
 
-instance Eq a => EqSequence [a] where
-    splitSeq = List.splitOn
-    stripPrefix = List.stripPrefix
-    stripSuffix x y = fmap reverse (List.stripPrefix (reverse x) (reverse y))
-    group = List.group
-    isPrefixOf = List.isPrefixOf
-    isSuffixOf x y = List.isPrefixOf (List.reverse x) (List.reverse y)
-    isInfixOf = List.isInfixOf
-    delete = List.delete
-    deleteBy = List.deleteBy
-    {-# INLINE splitSeq #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
-    {-# INLINE delete #-}
-    {-# INLINE deleteBy #-}
+{-# RULES "strict ByteString splitElem" splitElem = splitElemStrictBS #-}
+{-# RULES "strict ByteString stripPrefix" stripPrefix = stripPrefixStrictBS #-}
+{-# RULES "strict ByteString stripSuffix" stripSuffix = stripSuffixStrictBS #-}
+{-# RULES "strict ByteString group" group = S.group #-}
+{-# RULES "strict ByteString isPrefixOf" isPrefixOf = S.isPrefixOf #-}
+{-# RULES "strict ByteString isSuffixOf" isSuffixOf = S.isSuffixOf #-}
+{-# RULES "strict ByteString isInfixOf" isInfixOf = S.isInfixOf #-}
 
-instance EqSequence S.ByteString where
-    splitElem sep s | S.null s = [S.empty]
-                    | otherwise = S.split sep s
-    stripPrefix x y
-        | x `S.isPrefixOf` y = Just (S.drop (S.length x) y)
-        | otherwise = Nothing
-    stripSuffix x y
-        | x `S.isSuffixOf` y = Just (S.take (S.length y - S.length x) y)
-        | otherwise = Nothing
-    group = S.group
-    isPrefixOf = S.isPrefixOf
-    isSuffixOf = S.isSuffixOf
-    isInfixOf = S.isInfixOf
-    {-# INLINE splitElem #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
+splitElemStrictBS :: Word8 -> S.ByteString -> [S.ByteString]
+splitElemStrictBS sep s
+  | S.null s = [S.empty]
+  | otherwise = S.split sep s
 
-instance EqSequence L.ByteString where
-    splitElem sep s | L.null s = [L.empty]
-                    | otherwise = L.split sep s
-    stripPrefix x y
-        | x `L.isPrefixOf` y = Just (L.drop (L.length x) y)
-        | otherwise = Nothing
-    stripSuffix x y
-        | x `L.isSuffixOf` y = Just (L.take (L.length y - L.length x) y)
-        | otherwise = Nothing
-    group = L.group
-    isPrefixOf = L.isPrefixOf
-    isSuffixOf = L.isSuffixOf
-    isInfixOf x y = L.unpack x `List.isInfixOf` L.unpack y
-    {-# INLINE splitElem #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
+stripPrefixStrictBS :: S.ByteString -> S.ByteString -> Maybe S.ByteString
+stripPrefixStrictBS x y
+    | x `S.isPrefixOf` y = Just (S.drop (S.length x) y)
+    | otherwise = Nothing
 
-instance EqSequence T.Text where
-    splitSeq sep | T.null sep = (:) T.empty . List.map singleton . T.unpack
-                 | otherwise = T.splitOn sep
-    stripPrefix = T.stripPrefix
-    stripSuffix = T.stripSuffix
-    group = T.group
-    isPrefixOf = T.isPrefixOf
-    isSuffixOf = T.isSuffixOf
-    isInfixOf = T.isInfixOf
-    {-# INLINE splitSeq #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
+stripSuffixStrictBS :: S.ByteString -> S.ByteString -> Maybe S.ByteString
+stripSuffixStrictBS x y
+    | x `S.isSuffixOf` y = Just (S.take (S.length y - S.length x) y)
+    | otherwise = Nothing
 
-instance EqSequence TL.Text where
-    splitSeq sep | TL.null sep = (:) TL.empty . List.map singleton . TL.unpack
-                 | otherwise = TL.splitOn sep
-    stripPrefix = TL.stripPrefix
-    stripSuffix = TL.stripSuffix
-    group = TL.group
-    isPrefixOf = TL.isPrefixOf
-    isSuffixOf = TL.isSuffixOf
-    isInfixOf = TL.isInfixOf
-    {-# INLINE splitSeq #-}
-    {-# INLINE stripPrefix #-}
-    {-# INLINE stripSuffix #-}
-    {-# INLINE group #-}
-    {-# INLINE isPrefixOf #-}
-    {-# INLINE isSuffixOf #-}
-    {-# INLINE isInfixOf #-}
+{-# RULES "lazy ByteString splitElem" splitElem = splitSeqLazyBS #-}
+{-# RULES "lazy ByteString stripPrefix" stripPrefix = stripPrefixLazyBS #-}
+{-# RULES "lazy ByteString stripSuffix" stripSuffix = stripSuffixLazyBS #-}
+{-# RULES "lazy ByteString group" group = L.group #-}
+{-# RULES "lazy ByteString isPrefixOf" isPrefixOf = L.isPrefixOf #-}
+{-# RULES "lazy ByteString isSuffixOf" isSuffixOf = L.isSuffixOf #-}
 
-instance Eq a => EqSequence (Seq.Seq a)
-instance Eq a => EqSequence (V.Vector a)
-instance (Eq a, U.Unbox a) => EqSequence (U.Vector a)
-instance (Eq a, VS.Storable a) => EqSequence (VS.Vector a)
+splitSeqLazyBS :: Word8 -> L.ByteString -> [L.ByteString]
+splitSeqLazyBS sep s
+  | L.null s = [L.empty]
+  | otherwise = L.split sep s
 
--- | A typeclass for sequences whose elements have the 'Ord' typeclass
-class (EqSequence seq, MonoFoldableOrd seq) => OrdSequence seq where
-    -- | Sort a ordered sequence.
-    --
-    -- @
-    -- > 'sort' [4,3,1,2]
-    -- [1,2,3,4]
-    -- @
-    sort :: seq -> seq
-    sort = fromList . sort . otoList
-    {-# INLINE sort #-}
+stripPrefixLazyBS :: L.ByteString -> L.ByteString -> Maybe L.ByteString
+stripPrefixLazyBS x y
+    | x `L.isPrefixOf` y = Just (L.drop (L.length x) y)
+    | otherwise = Nothing
 
-instance Ord a => OrdSequence [a] where
-    sort = V.toList . sort . V.fromList
-    {-# INLINE sort #-}
+stripSuffixLazyBS :: L.ByteString -> L.ByteString -> Maybe L.ByteString
+stripSuffixLazyBS x y
+    | x `L.isSuffixOf` y = Just (L.take (L.length y - L.length x) y)
+    | otherwise = Nothing
 
-instance OrdSequence S.ByteString where
-    sort = S.sort
-    {-# INLINE sort #-}
+{-# RULES "strict Text splitSeq" splitSeq = splitSeqStrictText #-}
+{-# RULES "strict Text replaceSeq" replaceSeq = replaceSeqStrictText #-}
+{-# RULES "strict Text stripPrefix" stripPrefix = T.stripPrefix #-}
+{-# RULES "strict Text stripSuffix" stripSuffix = T.stripSuffix #-}
+{-# RULES "strict Text group" group = T.group #-}
+{-# RULES "strict Text isPrefixOf" isPrefixOf = T.isPrefixOf #-}
+{-# RULES "strict Text isSuffixOf" isSuffixOf = T.isSuffixOf #-}
+{-# RULES "strict Text isInfixOf" isInfixOf = T.isInfixOf #-}
 
-instance OrdSequence L.ByteString
-instance OrdSequence T.Text
-instance OrdSequence TL.Text
-instance Ord a => OrdSequence (Seq.Seq a)
+splitSeqStrictText :: T.Text -> T.Text -> [T.Text]
+splitSeqStrictText sep
+    | T.null sep = (:) T.empty . List.map singleton . T.unpack
+    | otherwise = T.splitOn sep
 
-instance Ord a => OrdSequence (V.Vector a) where
-    sort = vectorSort
-    {-# INLINE sort #-}
+replaceSeqStrictText :: T.Text -> T.Text -> T.Text -> T.Text
+replaceSeqStrictText old new
+    | T.null old = T.intercalate new . splitSeqStrictText old
+    | otherwise = T.replace old new
 
-instance (Ord a, U.Unbox a) => OrdSequence (U.Vector a) where
-    sort = vectorSort
-    {-# INLINE sort #-}
+{-# RULES "lazy Text splitSeq" splitSeq = splitSeqLazyText #-}
+{-# RULES "lazy Text replaceSeq" replaceSeq = replaceSeqLazyText #-}
+{-# RULES "lazy Text stripPrefix" stripPrefix = TL.stripPrefix #-}
+{-# RULES "lazy Text stripSuffix" stripSuffix = TL.stripSuffix #-}
+{-# RULES "lazy Text group" group = TL.group #-}
+{-# RULES "lazy Text isPrefixOf" isPrefixOf = TL.isPrefixOf #-}
+{-# RULES "lazy Text isSuffixOf" isSuffixOf = TL.isSuffixOf #-}
+{-# RULES "lazy Text isInfixOf" isInfixOf = TL.isInfixOf #-}
 
-instance (Ord a, VS.Storable a) => OrdSequence (VS.Vector a) where
-    sort = vectorSort
-    {-# INLINE sort #-}
+splitSeqLazyText :: TL.Text -> TL.Text -> [TL.Text]
+splitSeqLazyText sep
+    | TL.null sep = (:) TL.empty . List.map singleton . TL.unpack
+    | otherwise = TL.splitOn sep
+
+replaceSeqLazyText :: TL.Text -> TL.Text -> TL.Text -> TL.Text
+replaceSeqLazyText old new
+    | TL.null old = TL.intercalate new . splitSeqLazyText old
+    | otherwise = TL.replace old new
+
+-- | Sort a ordered sequence.
+--
+-- @
+-- > 'sort' [4,3,1,2]
+-- [1,2,3,4]
+-- @
+sort :: (IsSequence seq, Ord (Element seq)) => seq -> seq
+sort = fromList . V.toList . vectorSort . V.fromList . otoList
+{-# INLINE [0] sort #-}
+
+{-# RULES "strict ByteString sort" sort = S.sort #-}
+{-# RULES "boxed Vector sort" forall (v :: V.Vector a). sort v = vectorSort v #-}
+{-# RULES "unboxed Vector sort" forall (v :: U.Unbox a => U.Vector a). sort v = vectorSort v #-}
+{-# RULES "storable Vector sort" forall (v :: VS.Storable a => VS.Vector a). sort v = vectorSort v #-}
 
 -- | A typeclass for sequences whose elements are 'Char's.
 class (IsSequence t, IsString t, Element t ~ Char) => Textual t where
@@ -1453,7 +1405,7 @@ class (IsSequence t, IsString t, Element t ~ Char) => Textual t where
     -- > 'unwords' ["abc","def","ghi"]
     -- "abc def ghi"
     -- @
-    unwords :: [t] -> t
+    unwords :: (Element seq ~ t, MonoFoldable seq) => seq -> t
 
     -- | Break up a textual sequence at newline characters.
     --
@@ -1470,7 +1422,7 @@ class (IsSequence t, IsString t, Element t ~ Char) => Textual t where
     -- > 'unlines' ["abc","def","ghi"]
     -- "abc\\ndef\\nghi"
     -- @
-    unlines :: [t] -> t
+    unlines :: (Element seq ~ t, MonoFoldable seq) => seq -> t
 
     -- | Convert a textual sequence to lower-case.
     --
@@ -1520,9 +1472,9 @@ class (IsSequence t, IsString t, Element t ~ Char) => Textual t where
 
 instance (c ~ Char) => Textual [c] where
     words = List.words
-    unwords = List.unwords
+    unwords = List.unwords . otoList
     lines = List.lines
-    unlines = List.unlines
+    unlines = List.unlines . otoList
     toLower = TL.unpack . TL.toLower . TL.pack
     toUpper = TL.unpack . TL.toUpper . TL.pack
     toCaseFold = TL.unpack . TL.toCaseFold . TL.pack
@@ -1536,9 +1488,9 @@ instance (c ~ Char) => Textual [c] where
 
 instance Textual T.Text where
     words = T.words
-    unwords = T.unwords
+    unwords = T.unwords . otoList
     lines = T.lines
-    unlines = T.unlines
+    unlines = T.unlines . otoList
     toLower = T.toLower
     toUpper = T.toUpper
     toCaseFold = T.toCaseFold
@@ -1552,9 +1504,9 @@ instance Textual T.Text where
 
 instance Textual TL.Text where
     words = TL.words
-    unwords = TL.unwords
+    unwords = TL.unwords . otoList
     lines = TL.lines
-    unlines = TL.unlines
+    unlines = TL.unlines . otoList
     toLower = TL.toLower
     toUpper = TL.toUpper
     toCaseFold = TL.toCaseFold
@@ -1581,3 +1533,69 @@ catMaybes = fmap fromJust . filter isJust
 sortOn :: (Ord o, SemiSequence seq) => (Element seq -> o) -> seq -> seq
 sortOn = sortBy . comparing
 {-# INLINE sortOn #-}
+
+-- | Lazy sequences containing strict chunks of data.
+--
+-- @since 1.0.0
+class (IsSequence lazy, IsSequence strict) => LazySequence lazy strict | lazy -> strict, strict -> lazy where
+    toChunks :: lazy -> [strict]
+    fromChunks :: [strict] -> lazy
+    toStrict :: lazy -> strict
+    fromStrict :: strict -> lazy
+
+instance LazySequence L.ByteString S.ByteString where
+    toChunks = L.toChunks
+    fromChunks = L.fromChunks
+    toStrict = S.concat . L.toChunks
+    fromStrict = L.fromChunks . return
+
+instance LazySequence TL.Text T.Text where
+    toChunks = TL.toChunks
+    fromChunks = TL.fromChunks
+    toStrict = TL.toStrict
+    fromStrict = TL.fromStrict
+
+-- | Synonym for 'fromList'
+--
+-- @since 1.0.0
+pack :: IsSequence seq => [Element seq] -> seq
+pack = fromList
+{-# INLINE pack #-}
+
+-- | Synonym for 'otoList'
+--
+-- @since 1.0.0
+unpack :: MonoFoldable mono => mono -> [Element mono]
+unpack = otoList
+{-# INLINE unpack #-}
+
+-- | Repack from one type to another, dropping to a list in the middle.
+--
+-- @repack = pack . unpack@.
+--
+-- @since 1.0.0
+repack :: (MonoFoldable a, IsSequence b, Element a ~ Element b) => a -> b
+repack = pack . unpack
+
+-- | Textual data which can be encoded to and decoded from UTF8.
+--
+-- @since 1.0.0
+class (Textual textual, IsSequence binary) => Utf8 textual binary | textual -> binary, binary -> textual where
+    -- | Encode from textual to binary using UTF-8 encoding
+    --
+    -- @since 1.0.0
+    encodeUtf8 :: textual -> binary
+    -- | Note that this function is required to be pure. In the case of
+    -- a decoding error, Unicode replacement characters must be used.
+    --
+    -- @since 1.0.0
+    decodeUtf8 :: binary -> textual
+instance (c ~ Char, w ~ Word8) => Utf8 [c] [w] where
+    encodeUtf8 = L.unpack . TL.encodeUtf8 . TL.pack
+    decodeUtf8 = TL.unpack . TL.decodeUtf8With lenientDecode . L.pack
+instance Utf8 T.Text S.ByteString where
+    encodeUtf8 = T.encodeUtf8
+    decodeUtf8 = T.decodeUtf8With lenientDecode
+instance Utf8 TL.Text L.ByteString where
+    encodeUtf8 = TL.encodeUtf8
+    decodeUtf8 = TL.decodeUtf8With lenientDecode
