@@ -2,6 +2,7 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE CPP #-}
 
 module Spec where
 
@@ -10,15 +11,14 @@ import Data.Containers
 import Data.Sequences
 import qualified Data.Sequence as Seq
 import qualified Data.NonNull as NN
-import Data.ByteVector
 import Data.Monoid (mempty, mconcat)
 import Data.Maybe (fromMaybe)
+import qualified Data.List as List
 
 import Test.Hspec
 import Test.Hspec.QuickCheck
 import Test.HUnit ((@?=))
 import Test.QuickCheck hiding (NonEmptyList(..))
-import qualified Test.QuickCheck as QC
 import qualified Test.QuickCheck.Modifiers as QCM
 
 import Data.Text (Text)
@@ -34,20 +34,21 @@ import qualified Data.Semigroup as SG
 import qualified Data.Map as Map
 import qualified Data.IntMap as IntMap
 import qualified Data.HashMap.Strict as HashMap
-import qualified Data.IntSet as IntSet
 import qualified Data.Set as Set
 import qualified Control.Foldl as Foldl
 
-import Control.Arrow (first, second)
+import Control.Arrow (second)
 import Control.Applicative
 import Control.Monad.Trans.Writer
 
-import Prelude (Bool (..), ($), IO, min, abs, Eq (..), (&&), fromIntegral, Ord (..), String, mod, Int, Integer, show,
-                return, asTypeOf, (.), Show, id, (+), succ, Maybe (..), (*), mod, map, flip, otherwise, (-), div, seq, maybe)
+import Prelude (Bool (..), ($), IO, Eq (..), fromIntegral, Ord (..), String, mod, Int, Integer, show,
+                return, asTypeOf, (.), Show, (+), succ, Maybe (..), (*), mod, map, flip, otherwise, (-), div, maybe)
 import qualified Prelude
 
-instance Arbitrary a => Arbitrary (NE.NonEmpty a) where
-    arbitrary = (NE.:|) <$> arbitrary <*> arbitrary
+newtype NonEmpty' a = NonEmpty' (NE.NonEmpty a)
+    deriving (Show, Eq)
+instance Arbitrary a => Arbitrary (NonEmpty' a) where
+    arbitrary = NonEmpty' <$> ((NE.:|) <$> arbitrary <*> arbitrary)
 
 -- | Arbitrary newtype for key-value pairs without any duplicate keys
 -- and is not empty
@@ -99,7 +100,7 @@ main = hspec $ do
 
     describe "osum" $ do
         prop "works on lists" $ \(Small x) (Small y) ->
-            y >= x ==> osum [x..y] @?= ((x + y) * (y - x + 1) `div` 2)
+            y >= x ==> osum [x..y] @?= ((x + y) * (y - x + 1) `div` (2 :: Int))
 
     describe "oproduct" $ do
         prop "works on lists" $ \(Positive x) (Positive y) ->
@@ -203,14 +204,14 @@ main = hspec $ do
 
     describe "NonNull" $ do
         describe "fromNonEmpty" $ do
-            prop "toMinList" $ \ne ->
+            prop "toMinList" $ \(NonEmpty' ne) ->
                 (NE.toList ne :: [Int]) @?= NN.toNullable (NN.toMinList ne)
 
         let -- | Type restricted 'NN.ncons'
             nconsAs :: IsSequence seq => Element seq -> [Element seq] -> seq -> NN.NonNull seq
             nconsAs x xs _ = NN.ncons x (fromList xs)
 
-            test :: (OrdSequence typ, Arbitrary (Element typ), Show (Element typ), Show typ, Eq typ, Eq (Element typ))
+            test :: (IsSequence typ, Ord (Element typ), Arbitrary (Element typ), Show (Element typ), Show typ, Eq typ, Eq (Element typ))
                  => String -> typ -> Spec
             test typ du = describe typ $ do
                 prop "head" $ \x xs ->
@@ -309,15 +310,15 @@ main = hspec $ do
                      in updateMap f k (mapFromListAs xs dummy)
                             @?= mapFromList (updateMap f k xs)
 
-                prop "updateWithKey" $ \(DuplPairs xs) k ->
+                prop "updateWithKey" $ \(DuplPairs xs) k' ->
                     let f k i = if i < 0 then Nothing else Just $ i * k
-                     in updateWithKey f k (mapFromListAs xs dummy)
-                            @?= mapFromList (updateWithKey f k xs)
+                     in updateWithKey f k' (mapFromListAs xs dummy)
+                            @?= mapFromList (updateWithKey f k' xs)
 
-                prop "updateLookupWithKey" $ \(DuplPairs xs) k ->
+                prop "updateLookupWithKey" $ \(DuplPairs xs) k' ->
                     let f k i = if i < 0 then Nothing else Just $ i * k
-                     in updateLookupWithKey f k (mapFromListAs xs dummy)
-                            @?= second mapFromList (updateLookupWithKey f k xs)
+                     in updateLookupWithKey f k' (mapFromListAs xs dummy)
+                            @?= second mapFromList (updateLookupWithKey f k' xs)
 
                 prop "alter" $ \(DuplPairs xs) k ->
                     let m = mapFromListAs xs dummy
@@ -374,6 +375,32 @@ main = hspec $ do
                 x2 = Foldl.purely ofoldlUnwrap Foldl.length xs
             x2 @?= x1
 
+    describe "Replacing" $ do
+        let test typ dummy = describe typ $ do
+                prop "replaceElem old new === omap (\\x -> if x == old then new else x)" $
+                    -- replace random element or any random value with random new value
+                    \x list new -> forAll (elements (x:list)) $ \old ->
+                    let seq' = fromListAs list dummy
+                    in replaceElem old new seq' @?= omap (\x -> if x == old then new else x) seq'
+#if MIN_VERSION_QuickCheck(2,8,0)
+                prop "replaceSeq old new === ointercalate new . splitSeq old" $
+                    -- replace random subsequence with random new sequence
+                    \list new -> forAll (sublistOf list) $ \old ->
+                    let [seq', old', new'] = map (`fromListAs` dummy) [list, old, new]
+                    in replaceSeq old' new' seq' @?= ointercalate new' (splitSeq old' seq')
+                prop "replaceSeq old old === id" $ \list -> forAll (sublistOf list) $ \old ->
+                    let [seq', old'] = map (`fromListAs` dummy) [list, old]
+                    in replaceSeq old' old' seq' @?= seq'
+#endif
+        test "List" ([] :: [Int])
+        test "Vector" (V.empty :: V.Vector Int)
+        test "Storable Vector" (VS.empty :: VS.Vector Int)
+        test "Unboxed Vector" (U.empty :: U.Vector Int)
+        test "Strict ByteString" S.empty
+        test "Lazy ByteString" L.empty
+        test "Strict Text" T.empty
+        test "Lazy Text" TL.empty
+
     describe "Sorting" $ do
         let test typ dummy = describe typ $ do
                 prop "sortBy" $ \input -> do
@@ -393,9 +420,9 @@ main = hspec $ do
     describe "Intercalate" $ do
         let test typ dummy = describe typ $ do
                 prop "intercalate === defaultIntercalate" $ \list lists ->
-                    let seq = fromListAs list dummy
+                    let seq' = fromListAs list dummy
                         seqs = map (`fromListAs` dummy) lists
-                    in intercalate seq seqs @?= defaultIntercalate seq seqs
+                    in ointercalate seq' seqs @?= fromList (List.intercalate list lists)
         test "List" ([] :: [Int])
         test "Vector" (V.empty :: V.Vector Int)
         test "Storable Vector" (VS.empty :: VS.Vector Int)
@@ -412,7 +439,7 @@ main = hspec $ do
                 prop "intercalate sep . splitSeq sep === id" $
                     \(fromList' -> sep) ->
                     \(mconcat . map (maybe sep fromList') -> xs) ->
-                    intercalate sep (splitSeq sep xs) @?= xs
+                    ointercalate sep (splitSeq sep xs) @?= xs
                 prop "splitSeq mempty xs === mempty : map singleton (otoList xs)" $
                     \input ->
                     splitSeq mempty (fromList' input) @?= mempty : map singleton input
@@ -421,7 +448,7 @@ main = hspec $ do
                     splitSeq sep mempty @?= [mempty]
                 prop "intercalate (singleton sep) . splitElem sep === id" $
                     \sep -> \(fromSepList sep -> xs) ->
-                    intercalate (singleton sep) (splitElem sep xs) @?= xs
+                    ointercalate (singleton sep) (splitElem sep xs) @?= xs
                 prop "length . splitElem sep === succ . length . filter (== sep)" $
                     \sep -> \(fromSepList sep -> xs) ->
                     olength (splitElem sep xs) @?= olength (filter (== sep) xs) + 1
@@ -443,19 +470,12 @@ main = hspec $ do
         test "Strict Text" T.empty
         test "Lazy Text" TL.empty
 
-    describe "Data.ByteVector" $ do
-        prop "toByteVector" $ \ws ->
-            (otoList . toByteVector . fromList $ ws) @?= ws
-
-        prop "fromByteVector" $ \ws ->
-            (otoList . fromByteVector . fromList $ ws) @?= ws
-
     describe "Other Issues" $ do
         it "#26 headEx on a list works" $
             headEx (1 : filter Prelude.odd [2,4..]) @?= (1 :: Int)
 
         it "#31 find doesn't infinitely loop on NonEmpty" $
-            find (== "a") ("a" NE.:| ["d","fgf"]) @?= Just "a"
+            find (== "a") ("a" NE.:| ["d","fgf"]) @?= Just ("a" :: String)
 
         it "#83 head on Seq works correctly" $ do
             headEx (Seq.fromList [1 :: Int,2,3]) @?= (1 :: Int)
